import collections
from datetime import datetime
from typing import Dict, get_type_hints, Union

from jinja2 import Template

from gqlpycgen.client import Client
from gqlpycgen.utils import json_dumps


MAX_DEPTH = 3
ISO_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def timestamp():
    return datetime.utcnow().strftime(ISO_FORMAT)


union_template = Template("""{{ pre_prefix }}{{ name }} {
{{ prefix }}{% for type_obj in types %}... on {{ type_obj.__name__ }} {{ '{' }}
{% for k, v in get_type_hints(type_obj.__init__).items() %}
{{ property_to_gql(type_obj, k, v, indent + 1, True) }}
{% endfor %}
{{ prefix }}{{ '}' }}
{% endfor %}
}
""")


def union_gql(union, name, indent):
    if indent > MAX_DEPTH:
        return ""
    prefix = ("    " * indent)
    pre_prefix = ("    " * (indent - 1))
    types = union.__args__[:-1]
    # TODO - known issue, for the case where two of the types have a property with the same name and the don't have the exact same type it blows up
    #   keep in mind String != String!
    # type_properties = dict()
    # for ty in types:
    #     props = dict()
    #     for k, v in get_type_hints(type_obj.__init__).items()
    #         props[k] = v
    #     type_properties[ty.__name__] = props
    return union_template.render(
        name=name,
        types=types,
        prefix=prefix,
        pre_prefix=pre_prefix,
        get_type_hints=get_type_hints,
        property_to_gql=property_to_gql,
        indent=indent,
    )


def property_to_gql(cls, name, value, indent, add_alias=False):
    if hasattr(value, "__origin__") and value.__origin__ is Union and len(value.__args__) > 2:
        # stupid hack, it seems python adds a NoneType as part of a Union for all
        return union_gql(value, name, indent + 1)
    if hasattr(value, "__args__"):
        value = value.__args__[0]
    if cls.TYPES[name].startswith("List["):
        value = value.__args__[0]
        if hasattr(value, "__args__"):
            value = value.__args__[0]
            if hasattr(value, "__args__"):
                value = value.__args__[0]
    if hasattr(value, 'gql'):
        return value.gql(name, indent + 1, None)
    else:
        return ("    " * indent) + name


class QueryBase(object):

    def __init__(self, client: Client):
        self.client = client

    def query(self, query: str, variables: Dict) -> Dict:
        gql_query = query
        try:
            results = self.client.execute(gql_query, variables)
            # TODO - handle error responses
            return results
        except Exception as err:
            print("-" * 80)
            print(query)
            print(json_dumps(variables))
            print(err)
            print("-" * 80)
            return {"status": "err", "message": str(err)}

    def prepare(self, cls, name, variables, var_types):
        if hasattr(cls, "gql"):
            gql = cls.gql(name, 1, variables)
            # TODO - need to make variable types match exactly, ID != ID! *sigh*
            args = ", ".join(["${}: {}".format(arg, var_types[arg].__name__) for arg in variables.keys()])
            if len(variables):
                return "query %s (%s) { %s }" % (name, args, gql)
            else:
                return "query { %s }" % (gql)
        else:
            return "query { %s }" % (name)


class MutationBase(QueryBase):

    def prepare(self, cls, name, variables, var_types):
        if hasattr(cls, "gql"):
            gql = cls.gql(name, 1, variables)
            args = ", ".join(["${}: {}".format(arg, var_types[arg].__name__) for arg in variables.keys()])
            if len(variables):
                return "mutation %s (%s) { %s }" % (name, args, gql)
            else:
                return "mutation { %s }" % (gql)
        else:
            return "mutation { %s }" % (name)


class ObjectBase(object):

    def to_json(self) -> Dict:
        result = {}
        for k in self.FIELDS:
            value = getattr(self, k)
            if isinstance(value, list):
                result[k] = value
                if len(value) and hasattr(value[0], "to_json"):
                    result[k] = [v.to_json() for v in value]
            elif hasattr(value, "to_json"):
                result[k] = value.to_json()
            elif value is not None:
                result[k] = value
        return result

    @classmethod
    def gql(cls, name: str, indent: int, variables: Dict) -> str:
        if indent > MAX_DEPTH:
            return ""
        bits = []
        prefix = ("    " * indent)
        pre_prefix = ("    " * (indent - 1))
        if variables and len(variables):
            bits.append(pre_prefix + name + "(" + ", ".join(["{}: ${}".format(arg, arg) for arg in variables.keys()]) + ") {")
        else:
            bits.append(pre_prefix + name + " {")
        hints = get_type_hints(cls.__init__)
        for k, v in hints.items():
            bits.append(property_to_gql(cls, k, v, indent))
        bits.append(pre_prefix + "}")
        return "\n".join(bits)

    @classmethod
    def from_json(cls, obj: Dict):
        if obj is None:
            return None
        args = {}
        hints = get_type_hints(cls.__init__)
        for k, v in hints.items():
            if k in obj:
                if cls.TYPES[k].startswith("List["):
                    values = obj[k]
                    v = v.__args__[0]
                    if hasattr(v, 'from_json'):
                        args[k] = []
                        for value in values:
                            args[k].append(v.from_json(value))
                    else:
                        args[k] = values
                elif hasattr(v, 'from_json'):
                    args[k] = v.from_json(obj[k])
                else:
                    args[k] = obj[k]
            else:
                args[k] = None
        return cls(**args)



from datetime import datetime, date
from enum import EnumMeta
import json


ISO_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
DATE_FORMAT = "%Y-%m-%d"


def now():
    return datetime.utcnow().strftime(ISO_FORMAT)


def decode(val):
    if val and hasattr(val, "decode"):
        return val.decode("utf8")
    return val


class CustomJsonEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, bytes):
            return obj.decode("utf8")
        if isinstance(obj, datetime):
            return obj.strftime(ISO_FORMAT)
        if isinstance(obj, date):
            return obj.strftime(DATE_FORMAT)
        if hasattr(obj, "to_dict"):
            to_json = getattr(obj, "to_dict")
            if callable(to_json):
                return to_json()
        if hasattr(obj, "value"):
            return obj.value
        return json.JSONEncoder.default(self, obj)


def json_dumps(data, pretty=False):
    if pretty:
        return json.dumps(data, cls=CustomJsonEncoder, indent=4, sort_keys=True)
    return json.dumps(data, cls=CustomJsonEncoder)



import json

from jinja2 import Template


LF = '\n'
LFLF = '\n\n'
LFTABTAB = '\n        '


def resolve_type(typeObj):
    if typeObj.get("kind") == "LIST":
        return 'List[{}]'.format(resolve_type(typeObj.get("ofType")))
    if typeObj.get("kind") == "NON_NULL":
        ofType = typeObj.get("ofType")
        if ofType.get("kind") == "LIST":
            # TODO - Not sure how to handle this one - I think the type system for this
            # library needs a rework/refactor so it can handle theses cases more gracefully
            return 'List[{}]'.format(resolve_type(ofType.get("ofType")))
        else:
            return required(typeObj.get("ofType").get("name"))
    elif typeObj.get("kind"):
        return typeObj.get("name")
    else:
        raise Warning('type {} is not understood'.format(typeObj))
        return None


def required(typeStr):
    return "{}__Required".format(typeStr)


class Field(object):

    def __init__(self, name, gtype):
        self.name = name
        self.gtype = gtype

    def __str__(self):
        return "{}: '{}'=None".format(self.name, self.gtype)


objectTemplate = Template("""class {{name}}(ObjectBase):
    FIELDS = [{% for name in field_names %}"{{ name }}", {% endfor %}]
    TYPES = {{ '{' }}{% for field in fields %}"{{ field.name }}": "{{ field.gtype }}"{% if not loop.last %}, {% endif %}{% endfor %}{{ '}' }}

    def __init__(self, {% for field in fields %}{{ field }}{% if not loop.last %}, {% endif %}{% endfor %}):{% for field in fields %}
        self.{{ field.name }}{% if py36plus: %}: '{{ field.gtype }}'{% endif %} = {{ field.name }}{% endfor %}


class {{name}}__Required({{name}}):
    pass



""")


class Object(object):

    def __init__(self, typeObj):
        self.name = typeObj.get("name")
        self.field_names = [field.get("name") for field in typeObj.get("fields")]
        self.fields = []
        self.generate_argument_list(typeObj.get("fields"))

    def list_dependencies(self):
        deps = set()
        for field in self.fields:
            gtype = field.gtype
            if gtype.startswith("List"):
                gtype = gtype[5:-1]
            if gtype not in ["String", "Boolean", "Float", "Int", "ID"]:
                deps.add(gtype)
        return list(deps)

    def toPython(self, py36plus=True):
        return objectTemplate.render(**self.__dict__, py36plus=py36plus)

    def generate_argument_list(self, fields):
        for field in fields:
            name = field.get("name")
            ftype = resolve_type(field.get("type"))
            if ftype:
                self.fields.append(Field(name, ftype))
            else:
                raise Warning('field {} on {}, type not understood'.format(name, self.name))


class InputObject(Object):

    def __init__(self, typeObj):
        self.name = typeObj.get("name")
        self.field_names = [field.get("name") for field in typeObj.get("inputFields")]
        self.fields = []
        self.generate_argument_list(typeObj.get("inputFields"))


class GraphEnum(object):

    def __init__(self, typeObj):
        self.name = typeObj.get("name")
        self.values = [value.get("name") for value in typeObj.get("enumValues")]

    def toPython(self, py36plus=True):
        props = ['    {value} = "{value}"'.format(value=value) for value in self.values]
        return """class {0}(Enum):
{1}

    def __str__(self):
        return str(self.value)


{0}__Required = {0}


""".format(self.name, LF.join(props))


class Union(object):

    def __init__(self, typeObj):
        self.name = typeObj.get("name")
        self.possible_types = [pt.get("name") for pt in typeObj.get("possibleTypes")]

    def toPython(self, py36plus=True):
        return """{0} = Union[{1}]\n{0}__Required = Union[{1}]\n""".format(self.name, ", ".join(self.possible_types))


class Scalar(object):

    def __init__(self, typeObj):
        self.name = typeObj.get("name")

    def toPython(self, py36plus=True):
        return "{0} = NewType('{0}', str)\n{0}__Required = NewType('{0}', str)\n".format(self.name)


queryMethodTemplate = Template("""
    def {{name}}(self, {% for arg in args %}{{ arg }}{% if not loop.last %}, {% endif %}{% endfor %}) -> {{ returns }}:
        args = [{% for arg in args %}"{{ arg }}"{% if not loop.last %}, {% endif %}{% endfor %}]
        variables = dict()
        var_types = dict()
{% for arg in args %}
        if {{ arg.name }} is not None:
            var_types["{{ arg.name }}"] = {{ arg.gtype }}
            if hasattr({{ arg.name }}, "to_json"):
                variables["{{ arg.name }}"] = {{ arg.name }}.to_json()
            else:
                variables["{{ arg.name }}"] = {{ arg.name }}
{% endfor %}
        query = self.prepare({{ returns }}, "{{ name }}", variables, var_types)
        results = self.query(query, variables)
        return {{ returns }}.from_json(results.get("{{ name }}"))
""")


class QueryMethod(object):

    def __init__(self, typeObj):
        self.name = typeObj.get("name")
        self.returns = resolve_type(typeObj.get("type"))
        self.args = []
        self.generate_argument_list(typeObj.get("args"))

    def generate_argument_list(self, fields):
        for field in fields:
            name = field.get("name")
            ftype = resolve_type(field.get("type"))
            if ftype:
                self.args.append(Field(name, ftype))
            else:
                raise Warning('arg {} on {}, type not understood'.format(name, self.name))

    def toPython(self, py36plus=True):
        return queryMethodTemplate.render(**self.__dict__)


class Query(object):

    def __init__(self, typeObj):
        self.methods = []
        for field in typeObj.get("fields"):
            if not field.get("name").startswith("_"):
                method = QueryMethod(field)
                self.methods.append(method)

    def toPython(self, py36plus=True):
        return """class Query(QueryBase):
{}
""".format(LF.join([method.toPython(py36plus=py36plus) for method in self.methods]))


class Mutation(object):

    def __init__(self, typeObj):
        self.methods = []
        for field in typeObj.get("fields"):
            if not field.get("name").startswith("_"):
                method = QueryMethod(field)
                self.methods.append(method)

    def toPython(self, py36plus=True):
        return """class Mutation(MutationBase):
{}
""".format(LF.join([method.toPython(py36plus=py36plus) for method in self.methods]))


def filter_enums(typeObjs):
    for typeObj in typeObjs:
        if not typeObj.get("name").startswith("_") and typeObj.get("kind") == "ENUM":
            yield typeObj


def filter_input_objects(typeObjs):
    for typeObj in typeObjs:
        if not typeObj.get("name").startswith("_") and typeObj.get("kind") == "INPUT_OBJECT":
            yield typeObj


def filter_objects(typeObjs):
    for typeObj in typeObjs:
        if not typeObj.get("name").startswith("_") and typeObj.get("kind") == "OBJECT" and typeObj.get("name") not in ["Query", "Mutation"]:
            yield typeObj


def filter_interfaces(typeObjs):
    for typeObj in typeObjs:
        if not typeObj.get("name").startswith("_") and typeObj.get("kind") == "INTERFACE":
            yield typeObj


def filter_unions(typeObjs):
    for typeObj in typeObjs:
        if not typeObj.get("name").startswith("_") and typeObj.get("kind") == "UNION":
            yield typeObj


def filter_scalars(typeObjs):
    for typeObj in typeObjs:
        if not typeObj.get("name").startswith("_") and typeObj.get("kind") == "SCALAR" and typeObj.get("name") not in ["String", "Int", "Boolean", "Float", "ID"]:
            yield typeObj



from collections import OrderedDict
from uuid import uuid4

import requests

import tenacity
import logging

from gqlpycgen.utils import json_dumps

logger = logging.getLogger(__name__)

exponential_retry = tenacity.retry(
    stop = tenacity.stop_after_attempt(4),
    wait = tenacity.wait_exponential(multiplier=0.2/2),
    before = tenacity.before_log(logger, logging.DEBUG),
    after = tenacity.after_log(logger, logging.DEBUG),
    before_sleep = tenacity.before_sleep_log(logger, logging.WARNING)
)


class FileUpload(object):

    def __init__(self):
        self.map = dict()
        self.list = []
        self.containsFiles = False

    def add_file(self, var_path, filename, file, mimetype="application/octet-stream"):
        self.containsFiles = True
        file_id = uuid4().hex
        self.map[file_id] = [var_path]
        self.list.append((file_id, (filename, file, mimetype, ), ))


class Client(object):

    def __init__(self, uri, accessToken=None, client_credentials=None):
        self.uri = uri
        self.accessToken = accessToken
        self.client_credentials = client_credentials
        self.headers = {'Content-Type': 'application/json'}
        self.headers_files = {}
        if self.accessToken:
            self.headers["Authorization"] = self.headers_files["Authorization"] = f'bearer {self.accessToken}'
        elif self.accessToken is None and self.client_credentials:
            try:
                auth_response = requests.post(client_credentials["token_uri"], {
                    "audience": client_credentials["audience"],
                    "grant_type": "client_credentials",
                    "client_id": client_credentials["client_id"],
                    "client_secret": client_credentials["client_secret"],
                })
                self.accessToken = auth_response.json().get('access_token')
                if self.accessToken:
                    self.headers["Authorization"] = self.headers_files["Authorization"] = f'bearer {self.accessToken}'
                else:
                    raise Exception("invalid client_credentials")
            except Exception as err:
                # TODO log this error
                raise Exception("invalid client_credentials")

    @exponential_retry
    def execute(self, query, variables=None, files=None):
        payload = OrderedDict({
            'query': query,
            'variables': variables or {},
        })
        if files and files.containsFiles:
            data = {
                'operations': json_dumps(payload),
                'map': json_dumps(files.map),
            }
            request = requests.post(self.uri, data=data, files=files.list, headers=self.headers_files)
        else:
            request = requests.post(self.uri, data=json_dumps(payload), headers=self.headers)
        request.raise_for_status()
        result = request.json()
        if "errors" in result:
            return result.get("errors")
        return result.get("data")



from gqlpycgen.client import Client

schema_load_gql = """
{
  __schema {
    types {
      name
      kind
      interfaces {
        name
      }
      enumValues {
        name
      }
      fields {
        name
        args {
          name
          type {
            name
            kind
            ofType {
              name
              kind
            }
          }
        }
        type {
          name
          kind
          ofType {
            name
            kind
            ofType {
              name
              kind
              ofType {
                name
                kind
              }
            }
          }
          interfaces {
            name
          }
          possibleTypes {
            name
            kind
          }
        }
      }
      inputFields {
        name
        type {
          name
          kind
          ofType {
            name
            kind
            ofType {
              name
              kind
              ofType {
                name
                kind
              }
            }
          }
          interfaces {
            name
          }
          possibleTypes {
            name
            kind
          }
        }
      }
      possibleTypes {
        name
        kind
      }
    }
    queryType {
      name
      interfaces {
        name
      }
      fields {
        name
        isDeprecated
        deprecationReason
        description
        type {
          name
          kind
          ofType {
            name
            kind
            ofType {
              name
              kind
              ofType {
                name
                kind
              }
            }
          }
        }
        args {
          name
          defaultValue
          description
          type {
            name
            kind
            ofType {
              name
              kind
              ofType {
                name
                kind
                ofType {
                  name
                  kind
                }
              }
            }
          }
        }
      }
    }
    mutationType {
      name
      interfaces {
        name
      }
      fields {
        name
        isDeprecated
        deprecationReason
        description
        type {
          name
          kind
          ofType {
            name
            kind
            ofType {
              name
              kind
              ofType {
                name
                kind
              }
            }
          }
        }
        args {
          name
          defaultValue
          description
          type {
            name
            kind
            ofType {
              name
              kind
              ofType {
                name
                kind
                ofType {
                  name
                  kind
                }
              }
            }
          }
        }
      }
    }
  }
}
"""


def load_remote_schema(uri):
    client = Client(uri)
    schema = client.execute(schema_load_gql)
    return schema.get("__schema")



from unittest import TestCase

from mock import patch, MagicMock

from gqlpycgen import client

CLIENT_CREDS = {
    'token_uri': 'test.com',
    'audience': 'the queen',
    'client_id': 'foo',
    'client_secret': 'bar',
}



class ClientTest(TestCase):

    def test_client_init1(self):
        c = client.Client(uri='test.com', accessToken='ajkdsfjaksfh')

        self.assertEqual(c.uri, 'test.com')
        self.assertEqual(c.accessToken, 'ajkdsfjaksfh')
        self.assertEqual(c.headers["Authorization"], 'bearer ajkdsfjaksfh')
        self.assertIsNone(c.client_credentials)

    @patch('requests.post')
    def test_client_init2(self, post_mk):
        c = client.Client(uri='test.com', client_credentials=CLIENT_CREDS)

        post_mk.assert_called_once_with('test.com', {
            "audience": 'the queen',
            "grant_type": "client_credentials",
            "client_id": 'foo',
            "client_secret": 'bar',
        })

    @patch('requests.post')
    def test_client_init3(self, post_mk):
        c = client.Client(uri='test.com')
        post_mk.assert_not_called()

    @patch('requests.post', return_value=None)
    def test_client_init4(self, post_mk):
        with self.assertRaises(Exception):
            c = client.Client(uri='test.com', client_credentials=CLIENT_CREDS)

    @patch('requests.post')
    def test_client_init5(self, post_mk):
        post_mk.return_value.json = MagicMock()
        post_mk.return_value.json.return_value.get.return_value = None
        with self.assertRaises(Exception):
            c = client.Client(uri='test.com', client_credentials=CLIENT_CREDS)

    @patch('requests.post')
    def test_client_exec1(self, post_mk):
        c = client.Client(uri='test.com', client_credentials=CLIENT_CREDS)
        c.execute



from gqlpycgen.models import GraphEnum, filter_enums, Object, filter_objects, \
    filter_interfaces, filter_unions, filter_scalars, Union, Scalar, \
    filter_input_objects, InputObject, Query, Mutation
from gqlpycgen.loader import load_remote_schema


def write_header(out):
    out.write("# this file is generated, do not modify\n")
    out.write("from enum import Enum\n")
    out.write("from typing import List, NewType, TypeVar, Union\n")
    out.write('\n')
    out.write('from gqlpycgen.api import QueryBase, ObjectBase, MutationBase\n')
    out.write('\n')
    out.write('\n')
    out.write("ID = NewType('ID', str)\n")
    out.write("ID__Required = NewType('ID!', str)\n")
    out.write("Int = NewType('Int', int)\n")
    out.write("Int__Required = NewType('Int!', int)\n")
    out.write("String = NewType('String', str)\n")
    out.write("String__Required = NewType('String!', str)\n")
    out.write("Float = NewType('Float', float)\n")
    out.write("Float__Required = NewType('Float!', float)\n")
    out.write("Boolean = NewType('Boolean', bool)\n")
    out.write("Boolean__Required = NewType('Boolean!', bool)\n")


def do_remote(uri, filename, py36plus=True):
    schema = load_remote_schema(uri)
    schema_types = schema.get("types")

    with open(filename, 'w') as out:
        # includes the imports required and the core scalars
        write_header(out)

        # Pull all the scalars, filter_scalars skips the core primitives since we cover that in the headers
        for scalar in filter_scalars(schema_types):
            obj = Scalar(scalar)
            out.write(obj.toPython(py36plus=py36plus))

        # for PEP8
        out.write('\n')
        out.write('\n')

        # Pull all the ENUMs, turns them into python ENUM objects
        for enumType in filter_enums(schema_types):
            enum = GraphEnum(enumType)
            out.write(enum.toPython(py36plus=py36plus))

        # Pull interfaces, treat them as objects since in python they are just objects
        for objType in filter_interfaces(schema_types):
            obj = Object(objType)
            out.write(obj.toPython(py36plus=py36plus))

        # Pull objects and make them python objects
        for objType in filter_objects(schema_types):
            obj = Object(objType)
            out.write(obj.toPython(py36plus=py36plus))

        # Pull input objects and make them python objects
        for objType in filter_input_objects(schema_types):
            obj = InputObject(objType)
            out.write(obj.toPython(py36plus=py36plus))

        # lastly pull the unions and define TypeVars for them
        for union in filter_unions(schema_types):
            obj = Union(union)
            out.write(obj.toPython(py36plus=py36plus))

        out.write('\n')
        out.write('\n')

        # Pull the Query object and generate the API object
        query = Query(schema.get("queryType"))
        out.write(query.toPython(py36plus=py36plus))

        out.write('\n')
        out.write('\n')

        mutation = Mutation(schema.get("mutationType"))
        out.write(mutation.toPython(py36plus=py36plus))

        # don't forget to flush
        out.flush()



import os
from setuptools import setup, find_packages

BASEDIR = os.path.dirname(os.path.abspath(__file__))
VERSION = open(os.path.join(BASEDIR, 'VERSION')).read().strip()

BASE_DEPENDENCIES = [
    'Jinja2>=2.10',
    'click>=6.7',
    'requests>=2.21.0',
    'tenacity>=5.1.1'
]

TEST_DEPENDENCIES = [
    'coverage',
    'mock',
    'nose',
    'rednose'
]

# allow setup.py to be run from any path
os.chdir(os.path.normpath(BASEDIR))

setup(
    name='wf-gqlpycgen',
    version=VERSION,
    packages=find_packages(),
    include_package_data=True,
    description='GraphQL Python CLient Generator',
    long_description='Reads a GQL schema and turns it into a python library.',
    url='https://github.com/Wildflowerschools/graphql-python-client-generator',
    author='optimuspaul',
    author_email='paul.decoursey@wildflowerschools.org',
    install_requires=BASE_DEPENDENCIES,
    tests_require = TEST_DEPENDENCIES,
    extras_require = {
        'test': TEST_DEPENDENCIES
    },
    entry_points={
        'console_scripts': [
            'gqlpycgen=gqlpycgencli:cli',
        ],
    }
)
